# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pytest

from raiwidgets import ModelAnalysisDashboard
from responsibleai._interfaces import (CausalData, CounterfactualData, Dataset,
                                       ErrorAnalysisData, ModelExplanationData)


class TestModelAnalysisDashboard:

    def validate_rai_dashboard_data(self, rai_widget):
        assert isinstance(
            rai_widget.input.dashboard_input.dataset,
            Dataset)
        assert isinstance(
            rai_widget.input.dashboard_input.modelExplanationData[0],
            ModelExplanationData)
        assert isinstance(
            rai_widget.input.dashboard_input.errorAnalysisData[0],
            ErrorAnalysisData)
        assert isinstance(
            rai_widget.input.dashboard_input.causalAnalysisData[0],
            CausalData)
        assert isinstance(
            rai_widget.input.dashboard_input.counterfactualData[0],
            CounterfactualData)

    def test_model_analysis_adult(
            self, tmpdir,
            create_rai_insights_object_classification_with_model):
        ri = create_rai_insights_object_classification_with_model
        with pytest.warns(
            DeprecationWarning,
            match="MODULE-DEPRECATION-WARNING: "
                  "ModelAnalysisDashboard in raiwidgets package is "
                  "deprecated."
                  "Please use ResponsibleAIDashboard instead."):
            widget = ModelAnalysisDashboard(ri)
        self.validate_rai_dashboard_data(widget)

        save_dir = tmpdir.mkdir('save-dir')
        ri.save(save_dir)
        ri_copy = ri.load(save_dir)

        widget_copy = ModelAnalysisDashboard(ri_copy)
        self.validate_rai_dashboard_data(widget_copy)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import numpy as np
import pandas as pd
import sklearn
from interpret.ext.blackbox import MimicExplainer
from interpret.ext.glassbox import LGBMExplainableModel
from interpret_community.common.constants import ModelTask
from sklearn.datasets import make_classification
from sklearn.model_selection import train_test_split

from raiwidgets import ExplanationDashboard


class TestExplanationDashboardDashboard:
    def test_explanation_dashboard_many_columns(self):
        X, y = make_classification(n_features=2000)

        # Split data into train and test
        X_train, X_test, y_train, y_test = train_test_split(X,
                                                            y,
                                                            test_size=0.2,
                                                            random_state=0)
        classes = np.unique(y_train).tolist()
        feature_names = ["col" + str(i) for i in list(range(X_train.shape[1]))]
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)
        knn = sklearn.neighbors.KNeighborsClassifier()
        knn.fit(X_train, y_train)

        model_task = ModelTask.Classification
        explainer = MimicExplainer(knn,
                                   X_train,
                                   LGBMExplainableModel,
                                   model_task=model_task)
        global_explanation = explainer.explain_global(X_test)

        ExplanationDashboard(explanation=global_explanation, model=knn,
                             dataset=X_test, true_y=y_test, classes=classes)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Explanation dashboard class."""

from flask import jsonify, request

from .dashboard import Dashboard
from .explanation_dashboard_input import ExplanationDashboardInput


class ExplanationDashboard(Dashboard):
    """The dashboard class, wraps the dashboard component.

    :param explanation: An object that represents an explanation.
    :type explanation: ExplanationMixin
    :param model: An object that represents a model.
        It is assumed that for the classification case
        flit has a method of predict_proba()
        returning the prediction probabilities for each
        class and for the regression case a method of predict()
        returning the prediction value.
    :type model: object
    :param dataset: A matrix of feature vector examples
        (# examples x # features),
        the same samples used to build the explanation.
        Overwrites any existing dataset on the explanation object.
        Must have fewer than 100000 rows and fewer than 1000 columns.
        Note dashboard may become slow or crash for more than 10000 rows.
    :type dataset: numpy.ndarray or list[][]
    :param true_y: The true labels for the provided dataset.
        Overwrites any existing dataset on the explanation object.
    :type true_y: numpy.ndarray or list[]
    :param classes: The class names.
    :type classes: numpy.ndarray or list[]
    :param features: Feature names.
    :type features: numpy.ndarray or list[]
    :param public_ip: Optional. If running on a remote vm,
        the external public ip address of the VM.
    :type public_ip: str
    :param port: The port to use on locally hosted service.
    :type port: int
    :param locale: The language in which user wants to load and access the
        Explanation Dashboard. The default language is english ("en").
    :type locale: str
    """
    def __init__(self, explanation, model=None, dataset=None,
                 true_y=None, classes=None, features=None,
                 public_ip=None, port=None, locale=None):
        """Initialize the ExplanationDashboard."""
        self.input = ExplanationDashboardInput(
            explanation, model, dataset, true_y, classes, features)

        super(ExplanationDashboard, self).__init__(
            dashboard_type="Interpret",
            model_data=self.input.dashboard_input,
            public_ip=public_ip,
            port=port,
            locale=locale)

        def predict():
            data = request.get_json(force=True)
            return jsonify(self.input.on_predict(data))
        self.add_url_rule(predict, '/predict', methods=["POST"])



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Model Analysis Dashboard class."""

import warnings

from responsibleai import RAIInsights

from .responsibleai_dashboard import ResponsibleAIDashboard


class ModelAnalysisDashboard(object):
    """The dashboard class, wraps the dashboard component.

    Note: this class is now deprecated, please use the
    ResponsibleAIDashboard instead.

    :param analysis: An object that represents an model analysis.
    :type analysis: RAIInsights
    :param public_ip: Optional. If running on a remote vm,
        the external public ip address of the VM.
    :type public_ip: str
    :param port: The port to use on locally hosted service.
    :type port: int
    :param locale: The language in which user wants to load and access the
        ModelAnalysis Dashboard. The default language is english ("en").
    :type locale: str
    """
    def __init__(self, analysis: RAIInsights,
                 public_ip=None, port=None, locale=None):
        warnings.warn("MODULE-DEPRECATION-WARNING: "
                      "ModelAnalysisDashboard in raiwidgets package is "
                      "deprecated."
                      "Please use ResponsibleAIDashboard instead.",
                      DeprecationWarning)
        rai = ResponsibleAIDashboard(analysis, public_ip, port, locale)
        self.input = rai.input



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Model Analysis Dashboard class."""

from flask import jsonify, request

from raiwidgets.dashboard import Dashboard
from raiwidgets.responsibleai_dashboard_input import \
    ResponsibleAIDashboardInput
from responsibleai import RAIInsights


class ResponsibleAIDashboard(Dashboard):
    """The dashboard class, wraps the dashboard component.

    :param analysis: An object that represents an RAIInsights.
    :type analysis: RAIInsights
    :param public_ip: Optional. If running on a remote vm,
        the external public ip address of the VM.
    :type public_ip: str
    :param port: The port to use on locally hosted service.
    :type port: int
    :param locale: The language in which user wants to load and access the
        ResponsibleAI Dashboard. The default language is english ("en").
    :type locale: str
    :param cohort_list:
        List of cohorts defined by the user for the dashboard.
    :type cohort_list: List[Cohort]
    """
    def __init__(self, analysis: RAIInsights,
                 public_ip=None, port=None, locale=None,
                 cohort_list=None, **kwargs):
        self.input = ResponsibleAIDashboardInput(
            analysis, cohort_list=cohort_list)

        super(ResponsibleAIDashboard, self).__init__(
            dashboard_type="ResponsibleAI",
            model_data=self.input.dashboard_input,
            public_ip=public_ip,
            port=port,
            locale=locale,
            no_inline_dashboard=True,
            **kwargs)

        def predict():
            data = request.get_json(force=True)
            return jsonify(self.input.on_predict(data))
        self.add_url_rule(predict, '/predict', methods=["POST"])

        def tree():
            data = request.get_json(force=True)
            return jsonify(self.input.debug_ml(data))
        self.add_url_rule(tree, '/tree', methods=["POST"])

        def matrix():
            data = request.get_json(force=True)
            return jsonify(self.input.matrix(data))
        self.add_url_rule(matrix, '/matrix', methods=["POST"])

        def causal_whatif():
            data = request.get_json(force=True)
            return jsonify(self.input.causal_whatif(data))
        self.add_url_rule(causal_whatif, '/causal_whatif', methods=["POST"])

        def global_causal_effects():
            data = request.get_json(force=True)
            return jsonify(self.input.get_global_causal_effects(data))
        self.add_url_rule(global_causal_effects, '/global_causal_effects',
                          methods=["POST"])

        def global_causal_policy():
            data = request.get_json(force=True)
            return jsonify(self.input.get_global_causal_policy(data))
        self.add_url_rule(global_causal_policy, '/global_causal_policy',
                          methods=["POST"])

        def importances():
            return jsonify(self.input.importances())
        self.add_url_rule(importances, '/importances', methods=["POST"])

        def get_exp():
            data = request.get_json(force=True)
            return jsonify(self.input.get_exp(data))
        self.add_url_rule(get_exp, '/get_exp', methods=["POST"])

        def get_object_detection_metrics():
            data = request.get_json(force=True)
            return jsonify(self.input.get_object_detection_metrics(data))
        self.add_url_rule(
            get_exp,
            '/get_object_detection_metrics',
            methods=["POST"]
        )



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import json

import pandas as pd
import pytest

from raiwidgets.cohort import (ClassificationOutcomes, Cohort, CohortFilter,
                               CohortFilterMethods,
                               cohort_filter_json_converter)
from responsibleai.exceptions import UserConfigValidationException


def get_toy_binary_classification_dataset():
    return pd.DataFrame(data=[[23, 'X'], [25, 'Y']],
                        columns=["age", "target"])


def get_toy_multiclass_classification_dataset():
    return pd.DataFrame(
        data=[[23, 'X'], [25, 'Y'], [25, 'Z']],
        columns=["age", "target"])


def get_toy_regression_dataset():
    return pd.DataFrame(
        data=[[23, 2.5], [25, 3.6], [25, 4.6]],
        columns=["age", "target"])


class TestCohortFilter:
    def test_cohort_filter_validate_method(self):
        with pytest.raises(
                UserConfigValidationException,
                match="Got unexpected type <class 'int'> for method. "
                      "Expected string type."):
            CohortFilter(method=1,
                         arg=[], column=1)

        with pytest.raises(
                UserConfigValidationException,
                match="Got unexpected value random for method. "
                      "Expected either of greater or "
                      "greater and equal or "
                      "less or less and equal or "
                      "equal or includes or "
                      "excludes or in the range of."):
            CohortFilter(method="random", arg=[], column="random")

    def test_cohort_filter_validate_column(self):
        with pytest.raises(
                UserConfigValidationException,
                match="Got unexpected type <class 'int'> for column. "
                      "Expected string type."):
            CohortFilter(method=CohortFilterMethods.METHOD_GREATER,
                         arg=[], column=1)

    def test_cohort_filter_validate_arg(self):
        with pytest.raises(
                UserConfigValidationException,
                match="Got unexpected type <class 'int'> for arg. "
                      "Expected list type."):
            CohortFilter(method=CohortFilterMethods.METHOD_GREATER,
                         arg=1, column="age")

        with pytest.raises(
                UserConfigValidationException,
                match="Empty list supplied for arg."):
            CohortFilter(method=CohortFilterMethods.METHOD_GREATER,
                         arg=[], column="age")

    @pytest.mark.parametrize('method',
                             CohortFilterMethods.SINGLE_VALUE_METHODS)
    def test_cohort_filter_validate_single_value_methods_num_arg_entries(
            self, method):
        with pytest.raises(
                UserConfigValidationException,
                match="Expected a single value in arg "
                      "for cohort methods greater and "
                      "equal or greater or less and equal or less or equal."):
            CohortFilter(method=method, arg=[1, 9], column="age")

    @pytest.mark.parametrize('method',
                             CohortFilterMethods.SINGLE_VALUE_METHODS)
    def test_cohort_filter_validate_single_value_methods_type_arg_entries(
            self, method):
        with pytest.raises(
                UserConfigValidationException,
                match="Expected int or float type for arg "
                      "with cohort methods greater and "
                      "equal or greater or less and equal or less or equal."):
            CohortFilter(method=method, arg=["val"], column="age")

    def test_cohort_filter_validate_in_range_methods_num_arg_entries(
            self):
        with pytest.raises(
                UserConfigValidationException,
                match="Expected two entries in arg for "
                      "cohort method in the range of."):
            CohortFilter(method=CohortFilterMethods.METHOD_RANGE,
                         arg=[1], column="age")

    def test_cohort_filter_validate_in_range_methods_type_arg_entries(
            self):
        with pytest.raises(
                UserConfigValidationException,
                match="Expected int or float type for arg "
                      "with cohort method in the range of."):
            CohortFilter(method=CohortFilterMethods.METHOD_RANGE,
                         arg=[1, 'val'], column="age")

        with pytest.raises(
                UserConfigValidationException,
                match="Expected int or float type for arg "
                      "with cohort method in the range of."):
            CohortFilter(method=CohortFilterMethods.METHOD_RANGE,
                         arg=['val', 2], column="age")

    @pytest.mark.parametrize('method',
                             CohortFilterMethods.SINGLE_VALUE_METHODS)
    def test_cohort_filter_serialization_single_value_methods(self, method):
        cohort_filter_1 = CohortFilter(method=method,
                                       arg=[65.0], column='age')
        json_str = json.dumps(cohort_filter_1,
                              default=cohort_filter_json_converter)
        assert method in json_str
        assert '[65.0]' in json_str
        assert 'age' in json_str

    def test_cohort_filter_serialization_in_range_method(self):
        cohort_filter_1 = CohortFilter(
            method=CohortFilterMethods.METHOD_RANGE,
            arg=[65.0, 70.0], column='age')
        json_str = json.dumps(cohort_filter_1,
                              default=cohort_filter_json_converter)
        assert CohortFilterMethods.METHOD_RANGE in json_str
        assert '65.0' in json_str
        assert '70.0' in json_str
        assert 'age' in json_str

    @pytest.mark.parametrize('method',
                             [CohortFilterMethods.METHOD_INCLUDES,
                              CohortFilterMethods.METHOD_EXCLUDES])
    def test_cohort_filter_serialization_include_exclude_methods(self, method):
        cohort_filter_str = CohortFilter(method=method,
                                         arg=['val1', 'val2', 'val3'],
                                         column='age')
        json_str = json.dumps(cohort_filter_str,
                              default=cohort_filter_json_converter)
        assert method in json_str
        assert 'val1' in json_str
        assert 'val2' in json_str
        assert 'val3' in json_str
        assert 'age' in json_str

        cohort_filter_int = CohortFilter(method=method,
                                         arg=[1, 2, 3],
                                         column='age')
        json_str = json.dumps(cohort_filter_int,
                              default=cohort_filter_json_converter)
        assert method in json_str
        assert '1' in json_str
        assert '2' in json_str
        assert '3' in json_str
        assert 'age' in json_str


class TestCohortFilterDataValidations:
    def test_validate_with_test_data_high_level_validations(self):
        test_data = get_toy_binary_classification_dataset()

        cohort_filter_not_a_feature = CohortFilter(
            method=CohortFilterMethods.METHOD_LESS,
            arg=[65], column='fake_column')

        with pytest.raises(
                UserConfigValidationException,
                match="Unknown column fake_column specified in cohort filter"):
            cohort_filter_not_a_feature._validate_with_test_data(
                test_data=test_data, target_column="target",
                categorical_features=[])

    def test_validate_with_test_data_index_filter_validations(self):
        test_data = get_toy_binary_classification_dataset()

        cohort_filter_index_excludes = CohortFilter(
            method=CohortFilterMethods.METHOD_EXCLUDES,
            arg=[65], column=CohortFilter.INDEX)
        with pytest.raises(
                UserConfigValidationException,
                match="excludes filter is not supported with Index based "
                      "selection."):
            cohort_filter_index_excludes._validate_with_test_data(
                test_data=test_data, target_column="target",
                categorical_features=[]
            )

        cohort_filter_index_incorrect_args = CohortFilter(
            method=CohortFilterMethods.METHOD_GREATER,
            arg=[65.0], column=CohortFilter.INDEX)
        with pytest.raises(
                UserConfigValidationException,
                match="All entries in arg should be of type int."):
            cohort_filter_index_incorrect_args._validate_with_test_data(
                test_data=test_data, target_column="target",
                categorical_features=[]
            )

    def test_validate_with_test_data_classification_error_filter_validations(
            self):
        test_data_multiclass = get_toy_multiclass_classification_dataset()

        test_data_binary = get_toy_binary_classification_dataset()

        cohort_filter_classification_excludes = CohortFilter(
            method=CohortFilterMethods.METHOD_EXCLUDES,
            arg=[ClassificationOutcomes.FALSE_NEGATIVE],
            column=CohortFilter.CLASSIFICATION_OUTCOME)

        cohort_filter_classification_includes = CohortFilter(
            method=CohortFilterMethods.METHOD_INCLUDES,
            arg=["random"],
            column=CohortFilter.CLASSIFICATION_OUTCOME)

        with pytest.raises(
                UserConfigValidationException,
                match="Classification outcome cannot be "
                      "configured for multi-class classification"
                      " and regression scenarios."):
            cohort_filter_classification_excludes._validate_with_test_data(
                test_data=test_data_multiclass, target_column="target",
                categorical_features=[], is_classification=True
            )

        with pytest.raises(
                UserConfigValidationException,
                match="Classification outcome cannot be "
                      "configured for multi-class classification"
                      " and regression scenarios."):
            cohort_filter_classification_excludes._validate_with_test_data(
                test_data=test_data_binary, target_column="target",
                categorical_features=[], is_classification=False
            )

        with pytest.raises(
                UserConfigValidationException,
                match="Classification outcome can only be configured with "
                      "cohort filter includes."):
            cohort_filter_classification_excludes._validate_with_test_data(
                test_data=test_data_binary, target_column="target",
                categorical_features=[], is_classification=True
            )

        with pytest.raises(
                UserConfigValidationException,
                match="Classification outcome can only take argument values "
                      "from False negative or False positive or True "
                      "negative or True positive."):
            cohort_filter_classification_includes._validate_with_test_data(
                test_data=test_data_binary, target_column="target",
                categorical_features=[], is_classification=True)

    def test_validate_with_test_data_regression_error_filter_validations(
            self):
        test_data_regression = get_toy_regression_dataset()

        cohort_filter_regression = CohortFilter(
            method=CohortFilterMethods.METHOD_LESS,
            arg=[2.5],
            column=CohortFilter.REGRESSION_ERROR)

        with pytest.raises(
                UserConfigValidationException,
                match="Error cannot be configured for classification"
                      " scenarios."):
            cohort_filter_regression._validate_with_test_data(
                test_data=test_data_regression,
                target_column="target",
                categorical_features=[],
                is_classification=True)

        with pytest.raises(
                UserConfigValidationException,
                match="Error cannot be configured with either includes"
                      " or excludes."):
            cohort_filter_regression.method = \
                CohortFilterMethods.METHOD_INCLUDES
            cohort_filter_regression._validate_with_test_data(
                test_data=test_data_regression,
                target_column="target",
                categorical_features=[],
                is_classification=False)

        with pytest.raises(
                UserConfigValidationException,
                match="Error cannot be configured with either includes"
                      " or excludes."):
            cohort_filter_regression.method = \
                CohortFilterMethods.METHOD_EXCLUDES
            cohort_filter_regression._validate_with_test_data(
                test_data=test_data_regression,
                target_column="target",
                categorical_features=[],
                is_classification=False)

        with pytest.raises(
                UserConfigValidationException,
                match="All entries in arg should be of type int or float"
                      " for Error cohort."):
            cohort_filter_regression.method = \
                CohortFilterMethods.METHOD_GREATER
            cohort_filter_regression.arg = ['val1', 'val2']
            cohort_filter_regression._validate_with_test_data(
                test_data=test_data_regression,
                target_column="target",
                categorical_features=[],
                is_classification=False)

    @pytest.mark.parametrize('target_filter_type',
                             [CohortFilter.PREDICTED_Y,
                              CohortFilter.TRUE_Y])
    @pytest.mark.parametrize('method',
                             [CohortFilterMethods.METHOD_INCLUDES,
                              CohortFilterMethods.METHOD_EXCLUDES])
    def test_validate_with_test_data_regression_target_filter_validations(
            self, target_filter_type, method):
        test_data_regression = get_toy_regression_dataset()

        with pytest.raises(
                UserConfigValidationException,
                match="{0} cannot be configured with "
                      "filter {1} for regression.".format(target_filter_type,
                                                          method)):
            cohort_filter_regression = CohortFilter(
                method=method,
                arg=[2.5],
                column=target_filter_type)
            cohort_filter_regression._validate_with_test_data(
                test_data=test_data_regression,
                target_column="target",
                categorical_features=[],
                is_classification=False)

    @pytest.mark.parametrize('target_filter_type',
                             [CohortFilter.PREDICTED_Y,
                              CohortFilter.TRUE_Y])
    def test_validate_with_test_data_classification_target_filter_validations(
            self, target_filter_type):
        test_data_classification = get_toy_binary_classification_dataset()

        with pytest.raises(
                UserConfigValidationException,
                match="{0} can only be configured with "
                      "filter {1} for classification".format(
                          target_filter_type,
                          CohortFilterMethods.METHOD_INCLUDES)):
            cohort_filter_classification = CohortFilter(
                method=CohortFilterMethods.METHOD_EXCLUDES,
                arg=['X'],
                column=target_filter_type)
            cohort_filter_classification._validate_with_test_data(
                test_data=test_data_classification,
                target_column="target",
                categorical_features=[],
                is_classification=True)

        with pytest.raises(
            UserConfigValidationException,
            match="Found a class in arg which is not present in "
                  "test data"):
            cohort_filter_classification = CohortFilter(
                method=CohortFilterMethods.METHOD_INCLUDES,
                arg=['Z'],
                column=target_filter_type)
            cohort_filter_classification._validate_with_test_data(
                test_data=test_data_classification,
                target_column="target",
                categorical_features=[],
                is_classification=True)

    def test_validate_with_test_data_with_dataset_validations(
            self):
        test_data = pd.DataFrame(
            data=[[23, 'new', 'A'], [25, 'new, ''B'], [25, 'old', 'B']],
            columns=["age", 'type', "target"])

        with pytest.raises(
                UserConfigValidationException,
                match="{0} is a categorical feature and should be only "
                      "configured with {1} cohort filter.".format(
                          "type",
                          CohortFilterMethods.METHOD_INCLUDES)):
            cohort_filter = CohortFilter(
                method=CohortFilterMethods.METHOD_EXCLUDES,
                arg=['new'],
                column='type')
            cohort_filter._validate_with_test_data(
                test_data=test_data,
                target_column="target",
                categorical_features=['type'],
                is_classification=True)

        with pytest.raises(
                UserConfigValidationException,
                match="Found a category {0} in arg which is not present "
                      "in test data column {1}.".format('mid', 'type')):
            cohort_filter = CohortFilter(
                method=CohortFilterMethods.METHOD_INCLUDES,
                arg=['mid'],
                column='type')
            cohort_filter._validate_with_test_data(
                test_data=test_data,
                target_column="target",
                categorical_features=['type'],
                is_classification=True)


class TestCohort:
    def test_cohort_configuration_validations(self):
        with pytest.raises(
            UserConfigValidationException,
            match="Got unexpected type <class 'int'> for cohort name. "
                  "Expected string type."):
            Cohort(name=1)

        with pytest.raises(
            UserConfigValidationException,
            match="Got unexpected type <class 'list'> for cohort filter. "
                  "Expected CohortFilter type"):
            cohort = Cohort(name="Cohort New")
            cohort.add_cohort_filter(cohort_filter=[])

    def test_cohort_validate_with_test_data(self):
        cohort_filter_1 = CohortFilter(
            method=CohortFilterMethods.METHOD_LESS,
            arg=[65], column='age')
        cohort_1 = Cohort(name="Cohort New")
        cohort_1.add_cohort_filter(cohort_filter_1)
        test_data = get_toy_binary_classification_dataset()

        with pytest.raises(
                UserConfigValidationException,
                match="The test_data should be a pandas DataFrame"):
            cohort_1._validate_with_test_data(
                test_data=[], target_column='target',
                categorical_features=[])

        with pytest.raises(
                UserConfigValidationException,
                match="The target_column should be string."):
            cohort_1._validate_with_test_data(
                test_data=test_data,
                target_column=1,
                categorical_features=[])

        with pytest.raises(
                UserConfigValidationException,
                match="The target_column fake_target "
                      "was not found in test_data."):
            cohort_1._validate_with_test_data(
                test_data=test_data,
                target_column="fake_target",
                categorical_features=[])

        with pytest.raises(
                UserConfigValidationException,
                match="Expected a list type for "
                      "categorical columns."):
            cohort_1._validate_with_test_data(
                test_data=test_data,
                target_column="target",
                categorical_features={})

        with pytest.raises(
                UserConfigValidationException,
                match="Feature 1 in categorical_features need to be of "
                      "string type."):
            cohort_1._validate_with_test_data(
                test_data=test_data,
                target_column="target",
                categorical_features=[1, 2])

        with pytest.raises(
                UserConfigValidationException,
                match="Found categorical feature hours-per-week which is not"
                      " present in test data."):
            cohort_1._validate_with_test_data(
                test_data=test_data,
                target_column="target",
                categorical_features=["hours-per-week"])

    @pytest.mark.parametrize('method',
                             CohortFilterMethods.SINGLE_VALUE_METHODS)
    def test_cohort_serialization_single_value_method(self, method):
        cohort_filter_1 = CohortFilter(method=method,
                                       arg=[65], column='age')
        cohort_1 = Cohort(name="Cohort New")
        cohort_1.add_cohort_filter(cohort_filter_1)
        json_str = cohort_1.to_json()

        assert 'Cohort New' in json_str
        assert method in json_str
        assert '[65]' in json_str
        assert 'age' in json_str

    def test_cohort_serialization_deserialization_in_range_method(self):
        cohort_filter_1 = CohortFilter(
            method=CohortFilterMethods.METHOD_RANGE,
            arg=[65.0, 70.0], column='age')
        cohort_1 = Cohort(name="Cohort New")
        cohort_1.add_cohort_filter(cohort_filter_1)

        json_str = cohort_1.to_json()
        assert 'Cohort New' in json_str
        assert CohortFilterMethods.METHOD_RANGE in json_str
        assert '65.0' in json_str
        assert '70.0' in json_str
        assert 'age' in json_str

        cohort_1_new = Cohort.from_json(json_str)
        assert cohort_1_new.name == cohort_1.name
        assert len(cohort_1_new.cohort_filter_list) == \
            len(cohort_1.cohort_filter_list)
        assert cohort_1_new.cohort_filter_list[0].method == \
            cohort_1.cohort_filter_list[0].method

    @pytest.mark.parametrize('method',
                             [CohortFilterMethods.METHOD_INCLUDES,
                              CohortFilterMethods.METHOD_EXCLUDES])
    def test_cohort_serialization_deserialization_include_exclude_methods(
            self, method):
        cohort_filter_str = CohortFilter(method=method,
                                         arg=['val1', 'val2', 'val3'],
                                         column='age')
        cohort_str = Cohort(name="Cohort New Str")
        cohort_str.add_cohort_filter(cohort_filter_str)

        json_str = cohort_str.to_json()
        assert method in json_str
        assert 'val1' in json_str
        assert 'val2' in json_str
        assert 'val3' in json_str
        assert 'age' in json_str
        cohort_str_new = Cohort.from_json(json_str)
        assert cohort_str == cohort_str_new

        cohort_filter_int = CohortFilter(method=method,
                                         arg=[1, 2, 3],
                                         column='age')
        cohort_int = Cohort(name="Cohort New Int")
        cohort_int.add_cohort_filter(cohort_filter_int)

        json_str = cohort_int.to_json()
        assert method in json_str
        assert '1' in json_str
        assert '2' in json_str
        assert '3' in json_str
        assert 'age' in json_str

        cohort_int_new = Cohort.from_json(json_str)
        assert cohort_int == cohort_int_new

    def test_cohort_deserialization_error_conditions(self):
        test_dict = {}
        with pytest.raises(
                UserConfigValidationException,
                match="No name field found for cohort deserialization"):
            Cohort.from_json(json.dumps(test_dict))

        test_dict = {'name': 'Cohort New'}
        with pytest.raises(
                UserConfigValidationException,
                match="No cohort_filter_list field found for "
                      "cohort deserialization"):
            Cohort.from_json(json.dumps(test_dict))

        test_dict = {'name': 'Cohort New', 'cohort_filter_list': {}}
        with pytest.raises(UserConfigValidationException,
                           match="Field cohort_filter_list not of type list "
                                 "for cohort deserialization"):
            Cohort.from_json(json.dumps(test_dict))

        test_dict = {'name': 'Cohort New', 'cohort_filter_list': [{}]}
        with pytest.raises(
                UserConfigValidationException,
                match="No method field found for cohort filter "
                      "deserialization"):
            Cohort.from_json(json.dumps(test_dict))

        test_dict = {
            'name': 'Cohort New',
            'cohort_filter_list': [{"method": "fake_method"}]}
        with pytest.raises(
                UserConfigValidationException,
                match="No arg field found for cohort filter deserialization"):
            Cohort.from_json(json.dumps(test_dict))

        test_dict = {
            'name': 'Cohort New',
            'cohort_filter_list': [{"method": "fake_method", "arg": []}]}
        with pytest.raises(
                UserConfigValidationException,
                match="No column field found for cohort filter "
                      "deserialization"):
            Cohort.from_json(json.dumps(test_dict))


class TestCohortList:
    def test_cohort_list_serialization(self):
        cohort_filter_1 = CohortFilter(
            method=CohortFilterMethods.METHOD_LESS,
            arg=[65], column='age')
        cohort_1 = Cohort(name="Cohort New")
        cohort_1.add_cohort_filter(cohort_filter_1)

        cohort_2 = Cohort(name="Cohort Old")
        cohort_2.add_cohort_filter(cohort_filter_1)

        cohort_list = [cohort_1, cohort_2]
        json_str = json.dumps(cohort_list,
                              default=cohort_filter_json_converter)

        assert 'Cohort Old' in json_str
        assert 'Cohort New' in json_str
        assert CohortFilterMethods.METHOD_LESS in json_str
        assert '[65]' in json_str
        assert 'age' in json_str

    def test_empty_cohort_list_serialization(self):
        json_str = json.dumps(None, default=cohort_filter_json_converter)
        assert 'null' in json_str



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import numpy as np
import pytest
from pytest import approx

from raiwidgets.fairness_metric_calculation import (compute_wilson_bounds,
                                                    false_negative_rate_wilson,
                                                    false_positive_rate_wilson,
                                                    mae_standard_normal,
                                                    mse_standard_normal,
                                                    precision_wilson,
                                                    recall_wilson,
                                                    true_negative_rate_wilson)


@pytest.fixture()
def sample_binary_data():
    return np.array([0, 1, 1, 1, 0, 1, 0, 1, 0, 0]),\
        np.array([0, 1, 1, 1, 1, 1, 1, 0, 0, 0])


@pytest.fixture()
def sample_continuous_data():
    return np.array([25, 36, 12, 10, 52, 64, 34, 36, 11, 17]),\
        np.array([37, 29, 20, 2, 12, 75, 53, 64, 23, 29])


def test_compute_wilson_bounds():
    assert compute_wilson_bounds(0.5, 100) == (0.4038, 0.5962)
    assert compute_wilson_bounds(0.7, 100) == (0.6042, 0.7811)


def test_recall_wilson(sample_binary_data):
    y_true, y_pred = sample_binary_data
    assert recall_wilson(y_true, y_pred) == (0.3755, 0.9638)


def test_precision_wilson(sample_binary_data):
    y_true, y_pred = sample_binary_data
    assert precision_wilson(y_true, y_pred) == (0.3, 0.9032)


def test_false_positive_rate_wilson(sample_binary_data):
    y_true, y_pred = sample_binary_data
    assert false_positive_rate_wilson(y_true, y_pred) == (0.1176, 0.7693)


def test_true_negative_rate_wilson(sample_binary_data):
    y_true, y_pred = sample_binary_data
    assert true_negative_rate_wilson(y_true, y_pred) == (0.2307, 0.8824)


def test_false_negative_rate_wilson(sample_binary_data):
    y_true, y_pred = sample_binary_data
    assert false_negative_rate_wilson(y_true, y_pred) == (0.0362, 0.6245)


def test_mse_standard_normal_binary(sample_binary_data):
    y_true, y_pred = sample_binary_data
    mse = mse_standard_normal(y_true, y_pred)
    assert mse == approx((0.0160, 0.5840), rel=1e-3, abs=1e-3)


def test_mae_standard_normal_binary(sample_binary_data):
    y_true, y_pred = sample_binary_data
    mae = mae_standard_normal(y_true, y_pred)
    assert mae == approx((0.0160, 0.5840), rel=1e-3, abs=1e-3)


def test_mse_standard_normal_continuous(sample_continuous_data):
    y_true, y_pred = sample_continuous_data
    mse = mse_standard_normal(y_true, y_pred)
    assert mse == approx((57.7926, 637.2074), rel=1e-3, abs=1e-3)


def test_mae_standard_normal_continuous(sample_continuous_data):
    y_true, y_pred = sample_continuous_data
    mae = mae_standard_normal(y_true, y_pred)
    assert mae == approx((9.4708, 21.9292), rel=1e-3, abs=1e-3)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import warnings

import numpy as np
import pandas as pd

from raiutils.data_processing import convert_to_list, serialize_json_safe
from raiutils.models import is_classifier

from .constants import ErrorMessages
from .error_handling import _format_exception
from .explanation_constants import (ExplanationDashboardInterface,
                                    WidgetRequestResponseConstants)

EXP_VIZ_ERR_MSG = ErrorMessages.EXP_VIZ_ERR_MSG


class ExplanationDashboardInput:
    def __init__(
            self,
            explanation,
            model,
            dataset,
            true_y,
            classes,
            features):
        """Initialize the Explanation Dashboard Input.

        :param explanation: An object that represents an explanation.
        :type explanation: ExplanationMixin
        :param model: An object that represents a model.
        It is assumed that for the classification case
            it has a method of predict_proba() returning
            the prediction probabilities for each
            class and for the regression case a method of predict()
            returning the prediction value.
        :type model: object
        :param dataset: A matrix of feature vector examples
        (# examples x # features), the same samples
            used to build the explanation.
            Will overwrite any set on explanation object already.
            Must have fewer than
            100000 rows and fewer than 1000 columns.
            Note dashboard may become slow or crash for more than 10000 rows.
        :type dataset: numpy.ndarray or list[][]
        :param true_y: The true labels for the provided dataset.
            Will overwrite any set on
            explanation object already.
        :type true_y: numpy.ndarray or list[]
        :param classes: The class names.
        :type classes: numpy.ndarray or list[]
        :param features: Feature names.
        :type features: numpy.ndarray or list[]
        """
        self._model = model
        self._is_classifier = is_classifier(model)
        self._dataframeColumns = None
        self.dashboard_input = {}
        # List of explanations, key of explanation type is "explanation_type"
        if explanation is not None:
            self._mli_explanations = explanation.data(-1)["mli"]
        else:
            self._mli_explanations = None
        local_explanation = self._find_first_explanation(
            ExplanationDashboardInterface.MLI_LOCAL_EXPLANATION_KEY)
        global_explanation = self._find_first_explanation(
            ExplanationDashboardInterface.MLI_GLOBAL_EXPLANATION_KEY)
        ebm_explanation = self._find_first_explanation(
            ExplanationDashboardInterface.MLI_EBM_GLOBAL_EXPLANATION_KEY)
        dataset_explanation = self._find_first_explanation(
            ExplanationDashboardInterface.MLI_EXPLANATION_DATASET_KEY)

        if explanation is not None and hasattr(explanation, 'method'):
            self.dashboard_input[
                ExplanationDashboardInterface.EXPLANATION_METHOD
            ] = explanation.method

        predicted_y = None
        feature_length = None
        if dataset_explanation is not None:
            if dataset is None:
                dataset = dataset_explanation[
                    ExplanationDashboardInterface.MLI_DATASET_X_KEY
                ]
            if true_y is None:
                true_y = dataset_explanation[
                    ExplanationDashboardInterface.MLI_DATASET_Y_KEY
                ]

        if isinstance(dataset, pd.DataFrame) and hasattr(dataset, 'columns'):
            self._dataframeColumns = dataset.columns
            self._dfdtypes = dataset.dtypes
        try:
            list_dataset = convert_to_list(dataset, EXP_VIZ_ERR_MSG)
        except Exception as ex:
            ex_str = _format_exception(ex)
            raise ValueError(
                "Unsupported dataset type, inner error: {}".format(ex_str))
        if dataset is not None and model is not None:
            try:
                predicted_y = model.predict(dataset)
            except Exception as ex:
                ex_str = _format_exception(ex)
                msg = "Model does not support predict method for given"
                "dataset type, inner error: {}".format(
                    ex_str)
                raise ValueError(msg)
            try:
                predicted_y = convert_to_list(predicted_y, EXP_VIZ_ERR_MSG)
            except Exception as ex:
                ex_str = _format_exception(ex)
                raise ValueError(
                    "Model prediction output of unsupported type,"
                    "inner error: {}".format(ex_str))
        if predicted_y is not None:
            self.dashboard_input[
                ExplanationDashboardInterface.PREDICTED_Y
            ] = predicted_y
        row_length = 0
        if list_dataset is not None:
            row_length, feature_length = np.shape(list_dataset)
            if row_length > 100000:
                raise ValueError(
                    "Exceeds maximum number of rows"
                    "for visualization (100000)")
            if feature_length > 1000:
                warnings.warn("Exceeds maximum number of features for"
                              " visualization (1000)."
                              " Please regenerate the"
                              " explanation using fewer features or"
                              " initialize the dashboard without"
                              " passing a dataset. Dashboard will"
                              " show limited view.")
            else:
                self.dashboard_input[
                    ExplanationDashboardInterface.TRAINING_DATA
                ] = serialize_json_safe(list_dataset)
            self.dashboard_input[
                ExplanationDashboardInterface.IS_CLASSIFIER
            ] = self._is_classifier

        local_dim = None

        if true_y is not None and len(true_y) == row_length:
            self.dashboard_input[
                ExplanationDashboardInterface.TRUE_Y
            ] = convert_to_list(true_y, EXP_VIZ_ERR_MSG)

        if local_explanation is not None:
            try:
                local_explanation["scores"] = convert_to_list(
                    local_explanation["scores"], EXP_VIZ_ERR_MSG)
                local_explanation["intercept"] = convert_to_list(
                    local_explanation["intercept"], EXP_VIZ_ERR_MSG)
                # We can ignore perf explanation data.
                # Note if it is added back at any point,
                # the numpy values will need to be converted to python,
                # otherwise serialization fails.
                local_explanation["perf"] = None
                self.dashboard_input[
                    ExplanationDashboardInterface.LOCAL_EXPLANATIONS
                ] = local_explanation
            except Exception as ex:
                ex_str = _format_exception(ex)
                raise ValueError(
                    "Unsupported local explanation type,"
                    "inner error: {}".format(ex_str))
            if list_dataset is not None:
                local_dim = np.shape(local_explanation["scores"])
                if len(local_dim) != 2 and len(local_dim) != 3:
                    raise ValueError(
                        "Local explanation expected to be a 2D or 3D list")
                if len(local_dim) == 2 and (local_dim[1] != feature_length or
                                            local_dim[0] != row_length):
                    raise ValueError(
                        "Shape mismatch: local explanation"
                        "length differs from dataset")
                if len(local_dim) == 3 and (local_dim[2] != feature_length or
                                            local_dim[1] != row_length):
                    raise ValueError(
                        "Shape mismatch: local explanation"
                        " length differs from dataset")
        if local_explanation is None and global_explanation is not None:
            try:
                global_explanation["scores"] = convert_to_list(
                    global_explanation["scores"], EXP_VIZ_ERR_MSG)
                if 'intercept' in global_explanation:
                    global_explanation["intercept"] = convert_to_list(
                        global_explanation["intercept"], EXP_VIZ_ERR_MSG)
                self.dashboard_input[
                    ExplanationDashboardInterface.GLOBAL_EXPLANATION
                ] = global_explanation
            except Exception as ex:
                ex_str = _format_exception(ex)
                raise ValueError("Unsupported global explanation type,"
                                 "inner error: {}".format(ex_str))
        if ebm_explanation is not None:
            try:
                self.dashboard_input[
                    ExplanationDashboardInterface.EBM_EXPLANATION
                ] = ebm_explanation
            except Exception as ex:
                ex_str = _format_exception(ex)
                raise ValueError(
                    "Unsupported ebm explanation type: {}".format(ex_str))

        if features is None\
                and explanation is not None\
                and hasattr(explanation, 'features')\
                and explanation.features is not None:
            features = explanation.features
        if features is not None:
            features = convert_to_list(features, EXP_VIZ_ERR_MSG)
            if feature_length is not None and len(features) != feature_length:
                raise ValueError("Feature vector length mismatch:"
                                 " feature names length differs"
                                 " from local explanations dimension")
            self.dashboard_input[
                ExplanationDashboardInterface.FEATURE_NAMES
            ] = features
        if classes is None\
                and explanation is not None\
                and hasattr(explanation, 'classes')\
                and explanation.classes is not None:
            classes = explanation.classes
        if classes is not None:
            classes = convert_to_list(classes, EXP_VIZ_ERR_MSG)
            if local_dim is not None and len(classes) != local_dim[0]:
                raise ValueError("Class vector length mismatch:"
                                 "class names length differs from"
                                 "local explanations dimension")
            self.dashboard_input[
                ExplanationDashboardInterface.CLASS_NAMES
            ] = classes
        if is_classifier(model) and dataset is not None:
            try:
                probability_y = model.predict_proba(dataset)
            except Exception as ex:
                ex_str = _format_exception(ex)
                raise ValueError("Model does not support predict_proba method"
                                 " for given dataset type,"
                                 " inner error: {}".format(ex_str))
            try:
                probability_y = convert_to_list(probability_y,
                                                EXP_VIZ_ERR_MSG)
            except Exception as ex:
                ex_str = _format_exception(ex)
                raise ValueError(
                    "Model predict_proba output of unsupported type,"
                    "inner error: {}".format(ex_str))
            self.dashboard_input[
                ExplanationDashboardInterface.PROBABILITY_Y
            ] = probability_y

    def on_predict(self, data):
        try:
            if self._dataframeColumns is not None:
                data = pd.DataFrame(data, columns=self._dataframeColumns)
                data = data.astype(dict(self._dfdtypes))
            if (self._is_classifier):
                prediction = convert_to_list(
                    self._model.predict_proba(data), EXP_VIZ_ERR_MSG)
            else:
                prediction = convert_to_list(self._model.predict(data),
                                             EXP_VIZ_ERR_MSG)
            return {
                WidgetRequestResponseConstants.DATA: prediction
            }
        except Exception:
            return {
                WidgetRequestResponseConstants.ERROR: "Model threw exception"
                " while predicting...",
                WidgetRequestResponseConstants.DATA: []
            }

    def _find_first_explanation(self, key):
        if self._mli_explanations is None:
            return None
        new_array = [explanation for explanation
                     in self._mli_explanations
                     if explanation[
                         ExplanationDashboardInterface.MLI_EXPLANATION_TYPE_KEY
                     ] == key]
        if len(new_array) > 0:
            return new_array[0]["value"]
        return None



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the model performance dashboard class."""

from flask import jsonify, request

from .dashboard import Dashboard
from .explanation_dashboard_input import ExplanationDashboardInput


class ModelPerformanceDashboard(Dashboard):
    """The dashboard class, wraps the dashboard component.

    :param model: An object that represents a model.
        It is assumed that for the classification case
        flit has a method of predict_proba()
        returning the prediction probabilities for each
        class and for the regression case a method of predict()
        returning the prediction value.
    :type model: object
    :param dataset: A matrix of feature vector examples
        (# examples x # features),
        the same samples used to build the explanation.
        Overwrites any existing dataset on the explanation object.
        Must have fewer than 10000 rows and fewer than 1000 columns.
    :type dataset: numpy.ndarray or list[][]
    :param true_y: The true labels for the provided dataset.
        Overwrites any existing dataset on the explanation object.
    :type true_y: numpy.ndarray or list[]
    :param classes: The class names.
    :type classes: numpy.ndarray or list[]
    :param features: Feature names.
    :type features: numpy.ndarray or list[]
    :param public_ip: Optional. If running on a remote vm,
        the external public ip address of the VM.
    :type public_ip: str
    :param port: The port to use on locally hosted service.
    :type port: int
    :param locale: The language in which user wants to load and access the
        ModelPerformance Dashboard. The default language is english ("en").
    :type locale: str
    """
    def __init__(self, model=None, dataset=None,
                 true_y=None, classes=None, features=None,
                 public_ip=None, port=None, locale=None):
        """Initialize the model performance dashboard."""

        self.input = ExplanationDashboardInput(
            None, model, dataset, true_y, classes, features)

        Dashboard.__init__(self, dashboard_type="ModelPerformance",
                           model_data=self.input.dashboard_input,
                           public_ip=public_ip,
                           port=port,
                           locale=locale)

        def predict():
            data = request.get_json(force=True)
            return jsonify(self.input.on_predict(data))
        self.add_url_rule(predict, '/predict', methods=["POST"])



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Module for defining cohorts in raiwidgets package."""

import json
from typing import Any, List, Optional

import numpy as np
import pandas as pd

from responsibleai.exceptions import UserConfigValidationException


class CohortFilterMethods:
    """Defines different methods for cohort filters.
    """
    METHOD_GREATER = "greater"  # Strictly greater than a value
    # Greater than or equal to a value
    METHOD_GREATER_AND_EQUAL = "greater and equal"
    METHOD_LESS = "less"  # Strictly lesser than a value
    METHOD_LESS_AND_EQUAL = "less and equal"  # Lesser than or equal to a value
    METHOD_EQUAL = "equal"  # Equal to a value
    METHOD_INCLUDES = "includes"  # Includes a set of values
    METHOD_EXCLUDES = "excludes"  # Excludes a set of values
    METHOD_RANGE = "in the range of"  # In a given range of two values

    # List of all filter methods
    ALL = [METHOD_GREATER,
           METHOD_GREATER_AND_EQUAL,
           METHOD_LESS,
           METHOD_LESS_AND_EQUAL,
           METHOD_EQUAL,
           METHOD_INCLUDES,
           METHOD_EXCLUDES,
           METHOD_RANGE]

    # List of filter methods which expect a single value from the user
    SINGLE_VALUE_METHODS = [METHOD_GREATER_AND_EQUAL,
                            METHOD_GREATER,
                            METHOD_LESS_AND_EQUAL,
                            METHOD_LESS,
                            METHOD_EQUAL]


class ClassificationOutcomes:
    """Defines the possible values for classification outcomes.
    """
    FALSE_NEGATIVE = 'False negative'
    FALSE_POSITIVE = 'False positive'
    TRUE_NEGATIVE = 'True negative'
    TRUE_POSITIVE = 'True positive'

    ALL = [FALSE_NEGATIVE, FALSE_POSITIVE,
           TRUE_NEGATIVE, TRUE_POSITIVE]


def cohort_filter_json_converter(obj):
    """Helper function to convert CohortFilter object to json.
    :param obj: Object to convert to json.
    :type obj: object
    :return: The converted json.
    :rtype: dict
    """
    if isinstance(obj, CohortFilter):
        return obj.__dict__
    try:
        return obj.to_json()
    except AttributeError:
        return obj.__dict__


class CohortFilter:
    """Defines the cohort filter.
    :param method: Cohort filter method from one of CohortFilterMethods.
    :type method: str
    :param arg: List of values to be used by the cohort filter.
    :type arg: list
    :param column: The column name from the dataset on which the filter
                   will be applied.
    :type column: str
    """
    PREDICTED_Y = 'Predicted Y'
    TRUE_Y = 'True Y'
    INDEX = 'Index'
    CLASSIFICATION_OUTCOME = 'Classification outcome'
    REGRESSION_ERROR = 'Error'

    SPECIAL_COLUMN_LIST = [INDEX,
                           PREDICTED_Y,
                           TRUE_Y,
                           CLASSIFICATION_OUTCOME,
                           REGRESSION_ERROR]

    def __init__(self, method: str, arg: List[Any], column: str):
        """Defines the cohort filter.
        :param method: Cohort filter method from one of CohortFilterMethods.
        :type method: str
        :param arg: List of values to be used by the cohort filter.
        :type arg: list
        :param column: The column name from the dataset on which the filter
                       will be applied.
        :type column: str
        """
        self._validate_cohort_filter_parameters(
            method=method,
            arg=arg,
            column=column)

        self.method = method
        self.arg = arg
        self.column = column

    def __eq__(self, cohort_filter: Any):
        return self.method == cohort_filter.method and \
            self.arg == cohort_filter.arg and \
            self.column == cohort_filter.column

    def _validate_cohort_filter_parameters(
            self, method: str, arg: List[Any], column: str):
        """Validate the input values for the cohort filter.
        :param method: Cohort filter method from one of CohortFilterMethods.
        :type method: str
        :param arg: List of values to be used by the cohort filter.
        :type arg: list
        :param column: The column name from the dataset on which the filter
                       will be applied.
        :type column: str

        The following validations can be performed on the cohort filter:-

        1. Verify the correct types for method (expected string), column
           (expected string) and arg (expected list).
        2. The method value should be one of the filter string from
           CohortFilterMethods.ALL.
        3. The arg shouldn't be an empty list.
        4. For all cohort filter methods in
           CohortFilterMethods.SINGLE_VALUE_METHODS, the value in the arg
           should be integer or float and there should be only one value
           in arg.
        5. For cohort filter method CohortFilterMethods.METHOD_RANGE,
           the values in the arg should be integer or float and there
           should be only two values in arg.
        """
        if not isinstance(method, str):
            raise UserConfigValidationException(
                "Got unexpected type {0} for method. "
                "Expected string type.".format(type(method))
            )
        if method not in CohortFilterMethods.ALL:
            raise UserConfigValidationException(
                "Got unexpected value {0} for method. "
                "Expected either of {1}.".format(
                    method, " or ".join(CohortFilterMethods.ALL))
            )
        if not isinstance(column, str):
            raise UserConfigValidationException(
                "Got unexpected type {0} for column. "
                "Expected string type.".format(type(column))
            )
        if not isinstance(arg, list):
            raise UserConfigValidationException(
                "Got unexpected type {0} for arg. "
                "Expected list type.".format(type(arg))
            )
        if len(arg) == 0:
            raise UserConfigValidationException(
                "Empty list supplied for arg."
            )

        if method in CohortFilterMethods.SINGLE_VALUE_METHODS:
            if len(arg) != 1:
                raise UserConfigValidationException(
                    "Expected a single value in arg "
                    "for cohort methods {0}.".format(
                        " or ".join(CohortFilterMethods.SINGLE_VALUE_METHODS))
                )
            if not isinstance(arg[0], int) and not isinstance(arg[0], float):
                raise UserConfigValidationException(
                    "Expected int or float type for "
                    "arg with cohort methods {0}.".format(
                        " or ".join(CohortFilterMethods.SINGLE_VALUE_METHODS))
                )

        if method == CohortFilterMethods.METHOD_RANGE:
            if len(arg) != 2:
                raise UserConfigValidationException(
                    "Expected two entries in arg for "
                    "cohort method {0}.".format(
                        CohortFilterMethods.METHOD_RANGE)
                )
            if (not all(isinstance(entry, int) for entry in arg) and
                    not all(isinstance(entry, float) for entry in arg)):
                raise UserConfigValidationException(
                    "Expected int or float type for arg "
                    "with cohort method {0}.".format(
                        CohortFilterMethods.METHOD_RANGE)
                )

    def _validate_with_test_data(self, test_data: pd.DataFrame,
                                 target_column: str,
                                 categorical_features: List[str],
                                 is_classification: Optional[bool] = True):
        """
        Validate the cohort filters parameters with respect to test data.

        :param test_data: Test data over which cohort analysis will be done
            in ResponsibleAI Dashboard.
        :type test_data: pd.DataFrame
        :param target_column: The target column in the test data.
        :type target_column: str
        :param categorical_features: The categorical feature names.
        :type categorical_features: list[str]
        :param is_classification: True to indicate if this validation needs
            to be done for a classification scenario and False to indicate
            that this needs to be done for regression scenario.
        :type is_classification: bool

        The following validations need to be performed for cohort filter with
        test data:-

        High level validations
        1. Validate if the filter column is present in the test data.
        2. Validate if the filter column is present in the special column
           list.

        "Index" Filter validations
        1. The Index filter only takes integer arguments.
        2. The Index filter doesn't take CohortFilterMethods.EXCLUDES
           filter method.

        "Classification outcome" Filter validations
        1. Validate that "Classification outcome" filter is not configure for
           multiclass classification and regression.
        2. The "Classification outcome" filter only contains values from set
           ClassificationOutcomes.
        3. The "Classification outcome" filter only takes
           CohortFilterMethods.INCLUDES filter method.

        "Error" Filter validations
        1. Validate that "Error" filter is not configure for
           multiclass classification and binary classification.
        2. Only integer or floating points can be configured as arguments.
        3. The CohortFilterMethods.INCLUDES and CohortFilterMethods.EXCLUDES
           filter methods cannot be configured for this filter.

        "Predicted Y/True Y" Filter validations
        1. The set of classes configured in case of classification is a
           superset of the classes available in the test data.
        2. The CohortFilterMethods.INCLUDES is only allowed to be
           configured for "Predicted Y" filter in case of classification.
        3. The CohortFilterMethods.INCLUDES and CohortFilterMethods.EXCLUDES
           filter methods cannot be configured for this filter for regression.

        "Dataset" Filter validations
        1. TODO:- For continuous features the allowed values that be configured
           should be within the range of minimum and maximum values available
           within the continuous feature column in the test data.
        2. For categorical features only CohortFilterMethods.INCLUDES can be
           configured.
        3. For categorical features the values allowed are a subset of the
           the values available in the categorical column in the test data.
        """
        # High level validations
        if self.column not in CohortFilter.SPECIAL_COLUMN_LIST and \
                (self.column not in
                    (set(test_data.columns) - set([target_column]))):
            raise UserConfigValidationException(
                "Unknown column {0} specified in cohort filter".format(
                    self.column)
            )

        if self.column == CohortFilter.INDEX:
            # "Index" Filter validations
            if self.method == CohortFilterMethods.METHOD_EXCLUDES:
                raise UserConfigValidationException(
                    "{0} filter is not supported with {1} based "
                    "selection.".format(
                        CohortFilterMethods.METHOD_EXCLUDES,
                        CohortFilter.INDEX)
                )

            if not all(isinstance(entry, int) for entry in self.arg):
                raise UserConfigValidationException(
                    "All entries in arg should be of type int."
                )
        elif self.column == CohortFilter.CLASSIFICATION_OUTCOME:
            # "Classification outcome" Filter validations
            is_multiclass = len(np.unique(
                test_data[target_column].values).tolist()) > 2

            if not is_classification or is_multiclass:
                raise UserConfigValidationException(
                    "{0} cannot be configured for multi-class classification"
                    " and regression scenarios.".format(
                        CohortFilter.CLASSIFICATION_OUTCOME)
                )

            if self.method != CohortFilterMethods.METHOD_INCLUDES:
                raise UserConfigValidationException(
                    "{0} can only be configured with "
                    "cohort filter {1}.".format(
                        CohortFilter.CLASSIFICATION_OUTCOME,
                        CohortFilterMethods.METHOD_INCLUDES)
                )

            for classification_outcome in self.arg:
                if classification_outcome not in ClassificationOutcomes.ALL:
                    raise UserConfigValidationException(
                        "{0} can only take argument values from {1}.".format(
                            CohortFilter.CLASSIFICATION_OUTCOME,
                            " or ".join(ClassificationOutcomes.ALL))
                    )
        elif self.column == CohortFilter.REGRESSION_ERROR:
            # "Error" Filter validations
            if is_classification:
                raise UserConfigValidationException(
                    "{0} cannot be configured for classification"
                    " scenarios.".format(CohortFilter.REGRESSION_ERROR)
                )

            if self.method == CohortFilterMethods.METHOD_INCLUDES or \
                    self.method == CohortFilterMethods.METHOD_EXCLUDES:
                raise UserConfigValidationException(
                    "{0} cannot be configured with either {1} or {2}.".format(
                        CohortFilter.REGRESSION_ERROR,
                        CohortFilterMethods.METHOD_INCLUDES,
                        CohortFilterMethods.METHOD_EXCLUDES
                    )
                )

            if not all(isinstance(entry, int) for entry in self.arg) and \
                    not all(isinstance(entry, float) for entry in self.arg):
                raise UserConfigValidationException(
                    "All entries in arg should be of type int or float"
                    " for {} cohort.".format(CohortFilter.REGRESSION_ERROR)
                )
        elif self.column == CohortFilter.PREDICTED_Y or \
                self.column == CohortFilter.TRUE_Y:
            # "Predicted Y/True Y" Filter validations
            if is_classification:
                if self.method != CohortFilterMethods.METHOD_INCLUDES:
                    raise UserConfigValidationException(
                        "{0} can only be configured with "
                        "filter {1} for classification".format(
                            self.column,
                            CohortFilterMethods.METHOD_INCLUDES)
                    )

                test_classes = np.unique(
                    test_data[target_column].values).tolist()

                if not all(entry in test_classes for entry in self.arg):
                    raise UserConfigValidationException(
                        "Found a class in arg which is not present in "
                        "test data")
            else:
                if self.method == CohortFilterMethods.METHOD_INCLUDES or \
                        self.method == CohortFilterMethods.METHOD_EXCLUDES:
                    raise UserConfigValidationException(
                        "{0} cannot be configured with "
                        "filter {1} for regression.".format(
                            self.column, self.method)
                    )
        else:
            # "Dataset" Filter validations
            if self.column in categorical_features:
                if self.method != CohortFilterMethods.METHOD_INCLUDES:
                    raise UserConfigValidationException(
                        "{0} is a categorical feature and should be only "
                        "configured with {1} cohort filter.".format(
                            self.column,
                            CohortFilterMethods.METHOD_INCLUDES)
                    )

                categories = np.unique(
                    test_data[self.column].values).tolist()

                for entry in self.arg:
                    if entry not in categories:
                        raise UserConfigValidationException(
                            "Found a category {0} in arg which is not present "
                            "in test data column {1}.".format(
                                entry, self.column)
                        )


class Cohort:
    """Defines the cohort which will be injected from SDK into the Dashboard.
    :param name: Name of the cohort.
    :type name: str
    """
    def __init__(self, name: str):
        """Defines the cohort which will be injected from SDK into the Dashboard.
        :param name: Name of the cohort.
        :type name: str
        """
        if not isinstance(name, str):
            raise UserConfigValidationException(
                "Got unexpected type {0} for cohort name. "
                "Expected string type.".format(type(name))
            )
        self.name = name
        self.cohort_filter_list = None

    def __eq__(self, cohort: Any):
        same_name = self.name == cohort.name
        if self.cohort_filter_list is None and \
                cohort.cohort_filter_list is None:
            return same_name
        elif self.cohort_filter_list is not None and \
                cohort.cohort_filter_list is None:
            return False
        elif self.cohort_filter_list is None and \
                cohort.cohort_filter_list is not None:
            return False

        same_num_cohort_filters = len(self.cohort_filter_list) == \
            len(cohort.cohort_filter_list)
        if not same_num_cohort_filters:
            return False

        same_cohort_filters = True
        for index in range(0, len(self.cohort_filter_list)):
            if self.cohort_filter_list[index] != \
                    cohort.cohort_filter_list[index]:
                same_cohort_filters = False
                break

        return same_name and same_cohort_filters

    @staticmethod
    def _cohort_serializer(obj):
        """The function to serialize the Cohort class object.

        :param obj: Any member of the Cohort class object.
        :type: Any
        :return: Python dictionary.
        :rtype: dict[Any, Any]
        """
        return obj.__dict__

    def to_json(self):
        """Returns a serialized JSON string for the Cohort object.

        :return: The JSON string for the cohort.
        :rtype: str
        """
        return json.dumps(self, default=Cohort._cohort_serializer)

    @staticmethod
    def _get_cohort_object(json_dict):
        """Method to read a JSON dictionary and return a Cohort object.

        :param json_dict: JSON dictionary containing cohort data.
        :type: dict[str, str]
        :return: The Cohort object.
        :rtype: Cohort
        """
        cohort_fields = ["name", "cohort_filter_list"]
        for cohort_field in cohort_fields:
            if cohort_field not in json_dict:
                raise UserConfigValidationException(
                    "No {0} field found for cohort deserialization".format(
                        cohort_field))

        if not isinstance(json_dict['cohort_filter_list'], list):
            raise UserConfigValidationException(
                "Field cohort_filter_list not of type list for "
                "cohort deserialization")

        deserialized_cohort = Cohort(json_dict['name'])
        for serialized_cohort_filter in json_dict['cohort_filter_list']:
            cohort_filter_fields = ["method", "arg", "column"]
            for cohort_filter_field in cohort_filter_fields:
                if cohort_filter_field not in serialized_cohort_filter:
                    raise UserConfigValidationException(
                        "No {0} field found for cohort filter "
                        "deserialization".format(cohort_filter_field))

            cohort_filter = CohortFilter(
                method=serialized_cohort_filter['method'],
                arg=serialized_cohort_filter['arg'],
                column=serialized_cohort_filter['column'])
            deserialized_cohort.add_cohort_filter(cohort_filter=cohort_filter)
        return deserialized_cohort

    @staticmethod
    def from_json(json_str):
        """Method to read a json string and return a Cohort object.

        :param json_str: Serialized JSON string.
        :type: str
        :return: The Cohort object.
        :rtype: Cohort
        """
        json_dict = json.loads(json_str)
        return Cohort._get_cohort_object(json_dict)

    def add_cohort_filter(self, cohort_filter: CohortFilter):
        """Add a cohort filter into the cohort.
        :param cohort_filter: Cohort filter defined by CohortFilter class.
        :type: CohortFilter
        """
        if not isinstance(cohort_filter, CohortFilter):
            raise UserConfigValidationException(
                "Got unexpected type {0} for cohort filter. "
                "Expected CohortFilter type.".format(type(cohort_filter))
            )
        if self.cohort_filter_list is None:
            self.cohort_filter_list = [cohort_filter]
        else:
            self.cohort_filter_list.append(cohort_filter)

    def _validate_with_test_data(self, test_data: pd.DataFrame,
                                 target_column: str,
                                 categorical_features: List[str],
                                 is_classification: Optional[bool] = True):
        """
        Validate the cohort and cohort filters parameters with respect to
        test data.

        :param test_data: Test data over which cohort analysis will be done
            in ResponsibleAI Dashboard.
        :type test_data: pd.DataFrame
        :param target_column: The target column in the test data.
        :type target_column: str
        :param categorical_features: The categorical feature names.
        :type categorical_features: list[str]
        :param is_classification: True to indicate if this validation needs
            to be done for a classification scenario and False to indicate
            that this needs to be done for regression scenario.
        :type is_classification: bool
        """
        if self.cohort_filter_list is None:
            return
        if not isinstance(test_data, pd.DataFrame):
            raise UserConfigValidationException(
                "The test_data should be a pandas DataFrame.")
        if not isinstance(target_column, str):
            raise UserConfigValidationException(
                "The target_column should be string.")
        if not isinstance(categorical_features, list):
            raise UserConfigValidationException(
                "Expected a list type for categorical columns.")
        for categorical_feature in categorical_features:
            if not isinstance(categorical_feature, str):
                raise UserConfigValidationException(
                    "Feature {0} in categorical_features need to be of "
                    "string type.".format(categorical_feature)
                )

        if target_column not in test_data.columns:
            raise UserConfigValidationException(
                "The target_column {0} was not found in test_data.".format(
                    target_column)
            )

        test_data_columns_set = set(test_data.columns) - set([target_column])
        for categorical_feature in categorical_features:
            if categorical_feature not in test_data_columns_set:
                raise UserConfigValidationException(
                    "Found categorical feature {0} which is not"
                    " present in test data.".format(categorical_feature)
                )

        for cohort_filter in self.cohort_filter_list:
            cohort_filter._validate_with_test_data(
                test_data=test_data,
                target_column=target_column,
                categorical_features=categorical_features,
                is_classification=is_classification)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import traceback
from typing import List, Optional

import numpy as np
import pandas as pd

from erroranalysis._internal.constants import ModelTask, display_name_to_metric
from raiutils.data_processing import convert_to_list, serialize_json_safe
from raiutils.models import is_classifier
from raiwidgets.cohort import Cohort
from raiwidgets.constants import ErrorMessages
from raiwidgets.error_handling import _format_exception
from raiwidgets.interfaces import WidgetRequestResponseConstants
from responsibleai import RAIInsights
from responsibleai._internal.constants import ManagerNames
from responsibleai.exceptions import UserConfigValidationException

EXP_VIZ_ERR_MSG = ErrorMessages.EXP_VIZ_ERR_MSG


class ResponsibleAIDashboardInput:
    def __init__(
            self,
            analysis: RAIInsights,
            cohort_list: Optional[List[Cohort]] = None):
        """Initialize the Explanation Dashboard Input.

        :param analysis:
            A RAIInsights object that represents an explanation.
        :type analysis: RAIInsights
        :param cohort_list:
            List of cohorts defined by the user for the dashboard.
        :type cohort_list: List[Cohort]
        """
        self._analysis = analysis
        model = analysis.model
        self._is_classifier = is_classifier(model)
        self.dashboard_input = analysis.get_data()

        self._validate_cohort_list(cohort_list)
        if cohort_list is not None:
            # Add cohort_list to dashboard_input
            self.dashboard_input.cohortData = cohort_list
        else:
            self.dashboard_input.cohortData = []

        self._feature_length = len(self.dashboard_input.dataset.feature_names)
        if hasattr(analysis, ManagerNames.ERROR_ANALYSIS):
            self._error_analyzer = analysis.error_analysis._analyzer

    def _validate_cohort_list(self, cohort_list=None):
        if cohort_list is None:
            return

        if not isinstance(cohort_list, list):
            raise UserConfigValidationException(
                "cohort_list parameter should be a list.")

        if not all(isinstance(entry, Cohort) for entry in cohort_list):
            raise UserConfigValidationException(
                "All entries in cohort_list should be of type Cohort.")

        all_cohort_names = [cohort.name for cohort in cohort_list]
        unique_cohort_names = np.unique(all_cohort_names).tolist()

        if len(unique_cohort_names) != len(all_cohort_names):
            raise UserConfigValidationException(
                "Found cohorts with duplicate names. "
                "All pre-defined cohorts need to have distinct names.")

        test_data = pd.DataFrame(
            data=self.dashboard_input.dataset.features,
            columns=self.dashboard_input.dataset.feature_names)
        if self.dashboard_input.dataset.task_type == \
                ModelTask.CLASSIFICATION:
            class_names_list = self.dashboard_input.dataset.class_names
            true_y_array = self.dashboard_input.dataset.true_y
            true_class_array = np.array(
                [class_names_list[index] for index in true_y_array])
            test_data[self.dashboard_input.dataset.target_column] = \
                true_class_array
        else:
            test_data[self.dashboard_input.dataset.target_column] = \
                self.dashboard_input.dataset.true_y

        categorical_features = \
            self.dashboard_input.dataset.categorical_features
        for cohort in cohort_list:
            cohort._validate_with_test_data(
                test_data=test_data,
                target_column=self.dashboard_input.dataset.target_column,
                categorical_features=categorical_features,
                is_classification=self._is_classifier)

    def on_predict(self, data):
        try:
            data = pd.DataFrame(
                data, columns=self.dashboard_input.dataset.feature_names)
            data = self._analysis.get_test_data(test_data=data)
            if (self._is_classifier):
                prediction = convert_to_list(
                    self._analysis.model.predict_proba(data), EXP_VIZ_ERR_MSG)
            else:
                prediction = convert_to_list(
                    self._analysis.model.predict(data), EXP_VIZ_ERR_MSG)
            return {
                WidgetRequestResponseConstants.data: prediction
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error: "Model threw exception"
                " while predicting..."
                "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def _prepare_filtered_error_analysis_data(self, features, filters,
                                              composite_filters, metric):
        filtered_data_df = self._analysis.get_filtered_test_data(
            filters=filters,
            composite_filters=composite_filters,
            include_original_columns_only=False)

        msg = "Feature {} not found in dataset. Existing features: {}"
        for feature in features:
            if feature is None:
                continue
            if feature not in filtered_data_df.columns:
                raise UserConfigValidationException(
                    msg.format(feature, filtered_data_df.columns))

        self._error_analyzer.update_metric(metric)
        return filtered_data_df

    def debug_ml(self, data):
        try:
            features = data[0]
            filters = data[1]
            composite_filters = data[2]
            max_depth = data[3]
            num_leaves = data[4]
            min_child_samples = data[5]
            metric = display_name_to_metric[data[6]]

            filtered_data_df = self._prepare_filtered_error_analysis_data(
                features, filters, composite_filters, metric)

            tree = self._error_analyzer.compute_error_tree_on_dataset(
                features, filtered_data_df,
                max_depth, num_leaves, min_child_samples)
            return {
                WidgetRequestResponseConstants.data: tree
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to generate json tree representation,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def matrix(self, data):
        try:
            features = data[0]
            if features[0] is None and features[1] is None:
                return {WidgetRequestResponseConstants.data: []}
            filters = data[1]
            composite_filters = data[2]
            quantile_binning = data[3]
            num_bins = data[4]
            metric = display_name_to_metric[data[5]]

            filtered_data_df = self._prepare_filtered_error_analysis_data(
                features, filters, composite_filters, metric)

            matrix = self._error_analyzer.compute_matrix_on_dataset(
                features, filtered_data_df,
                quantile_binning, num_bins)
            return {
                WidgetRequestResponseConstants.data: matrix
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to generate json matrix representation,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def importances(self):
        try:
            scores = self._error_analyzer.compute_importances()
            return {
                WidgetRequestResponseConstants.data: scores
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to generate feature importances,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def causal_whatif(self, post_data):
        try:
            id, features, feature_name, new_value, target = post_data
            whatif = self._analysis.causal._whatif(
                id, self._analysis.get_test_data(
                    test_data=pd.DataFrame.from_records(features)), new_value,
                feature_name, target)
            return {
                WidgetRequestResponseConstants.data: whatif
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to generate causal what-if,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def get_exp(self, index):
        try:
            exp = self._analysis.explainer.compute_single_explanation(index)
            return {
                WidgetRequestResponseConstants.data: exp
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to generate image explanation,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def get_global_causal_effects(self, post_data):
        try:
            id = post_data[0]
            filters = post_data[1]
            composite_filters = post_data[2]
            filtered_data_df = self._analysis.get_filtered_test_data(
                filters=filters,
                composite_filters=composite_filters,
                include_original_columns_only=True)

            global_effects = \
                serialize_json_safe(
                    self._analysis.causal.request_global_cohort_effects(
                        id, filtered_data_df))
            return {
                WidgetRequestResponseConstants.data: global_effects
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to generate global causal effects for cohort,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def get_global_causal_policy(self, post_data):
        try:
            id = post_data[0]
            filters = post_data[1]
            composite_filters = post_data[2]
            filtered_data_df = self._analysis.get_filtered_test_data(
                filters=filters,
                composite_filters=composite_filters,
                include_original_columns_only=True)

            global_policy = \
                serialize_json_safe(
                    self._analysis.causal.request_global_cohort_policy(
                        id, filtered_data_df))
            return {
                WidgetRequestResponseConstants.data: global_policy
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to generate global causal policy for cohort,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }

    def get_object_detection_metrics(self, post_data):
        """Flask endpoint function to get Model Overview metrics
        for the Object Detection scenario.

        :param post_data: List of inputs in the order
        [true_y, predicted_y, aggregate_method, class_name, iou_thresh].
        :type post_data: List

        :return: JSON/dict data response
        :rtype: Dict[str, List]
        """
        try:
            true_y = post_data[0]
            predicted_y = post_data[1]
            aggregate_method = post_data[2]
            class_name = post_data[3]
            iou_thresh = post_data[4]
            exp = self._analysis.compute_object_detection_metrics(
                true_y,
                predicted_y,
                aggregate_method,
                class_name,
                iou_thresh
            )
            return {
                WidgetRequestResponseConstants.data: exp
            }
        except Exception as e:
            print(e)
            traceback.print_exc()
            e_str = _format_exception(e)
            return {
                WidgetRequestResponseConstants.error:
                    "Failed to get Object Detection Model Overview metrics,"
                    "inner error: {}".format(e_str),
                WidgetRequestResponseConstants.data: []
            }



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from unittest.mock import patch

import pytest

from raiwidgets import FairnessDashboard
from raiwidgets.fairness_metric_calculation import \
    MODULE_NOT_INSTALLED_ERROR_MESSAGE


@patch("importlib.import_module")
def test_no_fairlearn(importlib_mock):
    importlib_mock.side_effect = \
        ModuleNotFoundError("No module named 'fairlearn.metrics'")

    with pytest.raises(Exception) as exc:
        FairnessDashboard(
            sensitive_features={"a": [0, 1], "b": [0, 1]},
            y_true=[0, 1],
            y_pred=[0, 1])

    assert MODULE_NOT_INSTALLED_ERROR_MESSAGE.format('fairlearn.metrics') \
        in exc.value.args[0]



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import numpy as np
import pandas as pd
import shap
import sklearn
from interpret.ext.blackbox import MimicExplainer
from interpret.ext.glassbox import LGBMExplainableModel
from interpret_community.common.constants import ModelTask
from sklearn.datasets import load_iris, make_classification
from sklearn.model_selection import train_test_split

from erroranalysis._internal.constants import Metrics, metric_to_display_name
from erroranalysis._internal.surrogate_error_tree import (
    DEFAULT_MAX_DEPTH, DEFAULT_MIN_CHILD_SAMPLES, DEFAULT_NUM_LEAVES)
from raiwidgets import ErrorAnalysisDashboard
from raiwidgets.explanation_constants import WidgetRequestResponseConstants


class TestErrorAnalysisDashboard:
    def test_error_analysis_adult_census(self):
        X, y = shap.datasets.adult()
        y = [1 if r else 0 for r in y]
        categorical_features = ['Workclass',
                                'Education-Num',
                                'Marital Status',
                                'Occupation',
                                'Relationship',
                                'Race',
                                'Sex',
                                'Country']
        run_error_analysis_adult_census(X, y, categorical_features)

    def test_error_analysis_many_rows(self):
        X, y = make_classification(n_samples=110000)

        # Split data into train and test
        X_train, X_test, y_train, y_test = train_test_split(X,
                                                            y,
                                                            test_size=0.2,
                                                            random_state=0)
        classes = np.unique(y_train).tolist()
        feature_names = ["col" + str(i) for i in list(range(X_train.shape[1]))]
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)
        knn = sklearn.neighbors.KNeighborsClassifier()
        knn.fit(X_train, y_train)
        ErrorAnalysisDashboard(model=knn, dataset=X_test,
                               true_y=y_test, classes=classes)

    def test_error_analysis_sample_dataset_with_many_more_rows(self):
        X, y = make_classification(n_samples=400000)

        # Split data into train and test
        X_train, X_test, y_train, y_test = train_test_split(X,
                                                            y,
                                                            test_size=0.2,
                                                            random_state=0)
        classes = np.unique(y_train).tolist()
        feature_names = ["col" + str(i) for i in list(range(X_train.shape[1]))]
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)
        logreg = sklearn.linear_model.LogisticRegression()
        logreg.fit(X_train, y_train)
        _, X_test_sample, _, y_test_sample = train_test_split(X_test,
                                                              y_test,
                                                              test_size=0.01,
                                                              random_state=0)
        ErrorAnalysisDashboard(model=logreg,
                               dataset=X_test,
                               true_y=y_test_sample,
                               true_y_dataset=y_test,
                               classes=classes,
                               sample_dataset=X_test_sample)

        pred_y = logreg.predict(X_test)
        pred_y_sample = logreg.predict(X_test_sample)
        ErrorAnalysisDashboard(dataset=X_test,
                               true_y=y_test_sample,
                               true_y_dataset=y_test,
                               classes=classes,
                               sample_dataset=X_test_sample,
                               pred_y=pred_y_sample,
                               pred_y_dataset=pred_y)

    def test_error_analysis_pandas(self):
        X_train, X_test, y_train, y_test, feature_names, _ = create_iris_data()

        # Validate error analysis dashboard on pandas DataFrame
        # and pandas Series
        X_train = pd.DataFrame(X_train, columns=feature_names)
        X_test = pd.DataFrame(X_test, columns=feature_names)
        y_train = pd.Series(y_train)
        y_test = pd.Series(y_test)

        knn = sklearn.neighbors.KNeighborsClassifier()
        knn.fit(X_train, y_train)

        model_task = ModelTask.Classification
        explainer = MimicExplainer(knn,
                                   X_train,
                                   LGBMExplainableModel,
                                   model_task=model_task)
        global_explanation = explainer.explain_global(X_test)

        ErrorAnalysisDashboard(global_explanation,
                               knn,
                               dataset=X_test,
                               true_y=y_test)

    def test_error_analysis_iris_numeric_feature_names(self):
        # e2e test of error analysis with numeric feature names
        X_train, X_test, y_train, y_test, _, _ = create_iris_data()
        knn = sklearn.neighbors.KNeighborsClassifier()
        knn.fit(X_train, y_train)

        model_task = ModelTask.Classification
        explainer = MimicExplainer(knn,
                                   X_train,
                                   LGBMExplainableModel,
                                   model_task=model_task)
        global_explanation = explainer.explain_global(X_test)

        dashboard = ErrorAnalysisDashboard(global_explanation,
                                           knn,
                                           dataset=X_test,
                                           true_y=y_test)
        metric = metric_to_display_name[Metrics.ERROR_RATE]
        result = dashboard.input.debug_ml([global_explanation.features,
                                           [],
                                           [],
                                           DEFAULT_MAX_DEPTH,
                                           DEFAULT_NUM_LEAVES,
                                           DEFAULT_MIN_CHILD_SAMPLES,
                                           metric])
        assert WidgetRequestResponseConstants.ERROR not in result
        matrix_features = global_explanation.features[0:1]
        result = dashboard.input.matrix(matrix_features, [], [],
                                        True, 8, metric)
        assert WidgetRequestResponseConstants.ERROR not in result

    def test_error_analysis_adult_census_numeric_feature_names(self):
        X, y = shap.datasets.adult()
        categorical_features = ['Workclass',
                                'Education-Num',
                                'Marital Status',
                                'Occupation',
                                'Relationship',
                                'Race',
                                'Sex',
                                'Country']
        columns = X.columns.tolist()
        cat_idxs = [columns.index(feat) for feat in categorical_features]
        # Convert to numpy to remove features names
        X = X.values
        y = [1 if r else 0 for r in y]

        run_error_analysis_adult_census(X, y, cat_idxs)


def run_error_analysis_adult_census(X, y, categorical_features):
    X, y = sklearn.utils.resample(
        X, y, n_samples=1000, random_state=7, stratify=y)

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=7, stratify=y)

    knn = sklearn.neighbors.KNeighborsClassifier()
    knn.fit(X_train, y_train)

    model_task = ModelTask.Classification
    explainer = MimicExplainer(knn,
                               X_train,
                               LGBMExplainableModel,
                               augment_data=True,
                               max_num_of_augmentations=10,
                               model_task=model_task)
    global_explanation = explainer.explain_global(X_test)

    dashboard = ErrorAnalysisDashboard(
        global_explanation, knn, dataset=X_test,
        true_y=y_test, categorical_features=categorical_features)
    metric = metric_to_display_name[Metrics.ERROR_RATE]
    result = dashboard.input.debug_ml([global_explanation.features,
                                       [],
                                       [],
                                       DEFAULT_MAX_DEPTH,
                                       DEFAULT_NUM_LEAVES,
                                       DEFAULT_MIN_CHILD_SAMPLES,
                                       metric])
    assert WidgetRequestResponseConstants.ERROR not in result
    matrix_features = global_explanation.features[0:1]
    result = dashboard.input.matrix(matrix_features, [], [],
                                    True, 8, metric)
    assert WidgetRequestResponseConstants.ERROR not in result


def create_iris_data():
    # Import Iris dataset
    iris = load_iris()
    # Split data into train and test
    X_train, X_test, y_train, y_test = train_test_split(
        iris.data, iris.target, test_size=0.2, random_state=0)
    feature_names = iris.feature_names
    classes = iris.target_names
    return X_train, X_test, y_train, y_test, feature_names, classes
